unit Unit6;

interface

uses
  Winapi.Windows, Winapi.Messages, System.SysUtils, System.Variants, System.Classes, Vcl.Graphics,
  Vcl.Controls, Vcl.Forms, Vcl.Dialogs, Vcl.StdCtrls, IdBaseComponent, IdComponent, IdTCPConnection, IdTCPClient, IdHTTP, Vcl.ExtCtrls, IdIOHandler,
  IdIOHandlerSocket, IdIOHandlerStack, IdIOHandlerStream, IdHeaderList, Vcl.ComCtrls;

type
  TForm6 = class(TForm)
    bBajar: TButton;
    ECode: TEdit;
    Video: TIdHTTP;
    ProgressBar1: TProgressBar;
    Save: TSaveDialog;
    bCancelar: TButton;
    YouTube: TIdHTTP;
    Label1: TLabel;
    lProgress: TLabel;
    MSB: TIdHTTP;
    bCerrar: TButton;
    procedure bBajarClick(Sender: TObject);
    procedure VideoWork(ASender: TObject; AWorkMode: TWorkMode; AWorkCount: Int64);
    procedure bCancelarClick(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure bCerrarClick(Sender: TObject);
  private
    FCode: String;
    FURL: String;
    FName: String;
    FLength: Int64;
    FFormat: Integer;
    FChunkStart: Int64;

    function ValidateCode(const Code: String; GetURL: Boolean): String;
    procedure ParamsToStrings(const Params: String; StringList: TStrings);
    procedure GetYouTubeParams(const Response: String);
    procedure SaveCode;
    procedure RestoreCode;
    procedure SaveVideoDetails;
    function RestoreVideoDetails: Boolean;
    procedure CleanupCode;
  public
    { Public declarations }
  end;

var
  Form6: TForm6;

implementation

{$R *.dfm}

uses
  System.Types, StrUtils, Vcl.Imaging.pngimage, System.Win.Registry, Web.HTTPApp, IdUri, IdCoderMime;

const
  Reg: String = '\Software\MSBVideoDownloader';

// Ensure this function returns either a valid Youtube code, or a string prefixed with ERROR
function TForm6.ValidateCode(const Code: String; GetURL: Boolean): String;
var
  MS: TStringStream;
  Request: String;
  RequestType: String;
begin
  if Code = '' then raise Exception.Create('Ingrese su cdigo!');

  FCode := Code; // MSB code
  SaveCode;      // Saving until file successfully fully downloaded

  MS := TStringStream.Create;
  try
    if GetURL then RequestType := 'getcode' else RequestType := 'getid';
    Request := 'http://sanchezdebustamante.com/bajarvideo.php?'+RequestType+'&codigo='+String(HTTPEncode(AnsiString(FCode)));
    MSB.Get(Request, MS);
    Result := MS.DataString;
  except
    Result := 'ERROR: No se pudo acceder al validador del cdigo';
  end;
  MS.Free;
end;

procedure TForm6.bBajarClick(Sender: TObject);
var
  FS: TFileStream;
  Response: String;

  function ShowError: Boolean;
  begin
    Result := (Copy(Response, 1, 5) = 'ERROR');
    if Result then
    begin
      Response := Response.Replace('&oacute;', '').Replace('&iacute;', '').Replace('&aacute;', '').Replace('&eacute;', '').
        Replace('&uacute;', '').Replace('&ntilde;', '');
      ShowMessage(Response);
    end;
  end;
begin
  FCode := ECode.Text;

  if not RestoreVideoDetails then
  begin
    Response := ValidateCode(FCode, false);

    if ShowError then Exit;

    Response := TIdURI.URLDecode(Response);
    Response := TIdDecoderMIME.DecodeString(Response);

    GetYouTubeParams(Response);

    if FUrl = 'fail' then
    begin
      Response := ValidateCode(FCode, true);

      if ShowError then Exit;

      FUrl := Response;
      FName := 'Video.mp4';
      FLength := 0;
      FChunkStart := 0;
    end;
  end;

  if (Save.FileName = '') and (FName <> '') then
  begin
    FName := ReplaceStr(FName, '+', ' ');
    Save.FileName := FName;
    SaveVideoDetails;
  end;

  if not Save.Execute then Exit;

  bCancelar.Enabled := True;
  bCerrar.Enabled := False;
  bBajar.Enabled := False;

  if FileExists(Save.FileName) then
    FS := TFileStream.Create(Save.FileName, fmOpenReadWrite)
  else
    FS := TFileStream.Create(Save.FileName, fmCreate);

  try
    FS.Seek(0, soFromEnd);
    Video.Request.Ranges.Clear;

    if FS.Position > 0 then
    begin
      if (FLength > 0) and (FS.Position = FLength) then
      begin
        CleanupCode;
        ShowMessage('El video '+Save.FileName+' est completo!');
        Exit;
      end;
      Video.Request.Ranges.Add.StartPos := FS.Position;
    end;

    FChunkStart := FS.Position;

    Video.Get(FUrl, FS);

    FS.Seek(0, soFromEnd);
    if (FLength > 0) and (FS.Position = FLength) then
    begin
      CleanupCode;
      ShowMessage('El video '+Save.FileName+' ha bajado con xito.');
      Exit;
    end;

  finally
    FS.Free;
    bCancelar.Enabled := False;
    bCerrar.Enabled := True;
    bBajar.Enabled := True;
  end;
end;

procedure TForm6.bCancelarClick(Sender: TObject);
begin
  if Video.Connected then
    Video.Disconnect;
end;

procedure TForm6.bCerrarClick(Sender: TObject);
begin
  Close;
end;

procedure TForm6.CleanupCode;
var
  R: TRegistry;
begin
  FURL := '';
  FName := '';
  FLength := 0;
  FFormat := 0;
  FChunkStart := 0;

  R := TRegistry.Create;
  try
    R.DeleteKey(Reg+'\'+FCode);

    if R.OpenKey(Reg, False) then
      R.DeleteValue('LastCode');
  finally
    R.Free;

    FCode := '';
    ECode.Text := '';
  end;
end;

procedure TForm6.FormShow(Sender: TObject);
begin
  RestoreCode;
end;

procedure TForm6.GetYouTubeParams(const Response: String);
var
  YTCode: String;
  Params: String;
  ParamList: TStringDynArray;
  MS: TStringStream;
  TS: TStringList;
  N: Integer;
  UsedFormat: Integer;
begin
  if Pos(',', Response) > 0 then
  begin
    YTCode := Copy(Response, 1, Pos(',', Response)-1);
    FFormat := StrToInt(Copy(Response, Pos(',', Response)+1, 10));
  end else
  begin
    YTCode := Response;
    FFormat := 22; //default
  end;

  FUrl := '';
  FName := '';
  FLength := 0;
  FChunkStart := 0;

  MS := TStringStream.Create;
  TS := TStringList.Create;
  try
    YouTube.Get('http://www.youtube.com/get_video_info?video_id='+YTCode, MS);
    ParamsToStrings(MS.DataString, TS);
    if TS.Values['status'] = 'fail' then
    begin
      FUrl := 'fail';
      Exit;
    end;

    FName := TS.Values['title'];
    Params := TS.Values['url_encoded_fmt_stream_map'];
    if Params <> '' then
    begin
      UsedFormat := 0;
      ParamList := SplitString(Params, ',');
      for N := Low(ParamList) to High(ParamList) do
      begin
        ParamsToStrings(ParamList[N], TS);
        if TS.Values['itag'] = IntToStr(FFormat) then
        begin
          FURL := TS.Values['url'];
          UsedFormat := FFormat;
          Break;
        end else if (FURL = '') and TS.Values['type'].StartsWith('video/mp4') then
        begin
          FURL := TS.Values['url'];
          if TS.Values['itag'] <> '' then
            UsedFormat := StrToInt(TS.Values['itag']);
        end;
      end;
      if FUrl <> '' then
        FFormat := UsedFormat;
    end;

    SaveVideoDetails;
  finally
    TS.Free;
    MS.Free;
  end;
end;

procedure TForm6.ParamsToStrings(const Params: String; StringList: TStrings);
var
  ParamList: TStringDynArray;
  N: Integer;
begin
  StringList.Clear;

  if Params = '' then Exit;

  ParamList := SplitString(Params, '&');

  for N := Low(ParamList) to High(ParamList) do
    StringList.Add(TIdURI.URLDecode(ParamList[N]));
end;

procedure TForm6.RestoreCode;
var
  R: TRegistry;
begin
  R := TRegistry.Create;
  try
    if R.OpenKey(Reg, True) then
    begin
      FCode := R.ReadString('LastCode');
      if FCode > '' then
        ECode.Text := FCode;
    end;
  finally
    R.Free;
  end;
end;

function TForm6.RestoreVideoDetails: Boolean;
var
  R: TRegistry;
  Info: TRegDataInfo;
begin
  Result := False;

  if FCode = '' then Exit;

  R := TRegistry.Create;
  try
    if R.OpenKey(Reg+'\'+FCode, False) then
    begin
      FURL := R.ReadString('URL');
      FName := R.ReadString('Name');
      FFormat := R.ReadInteger('Format');

      if R.GetDataInfo('Length', Info) then
        if Info.RegData = rdBinary then
          R.ReadBinaryData('Length', FLength, sizeof(FLength));
    end;

    Result := (FURL <> '') and (FFormat > 0) and (FName <> '');
  finally
    R.Free;
  end;
end;

procedure TForm6.SaveCode;
var
  R: TRegistry;
begin
  FCode := Trim(ECode.Text);
  if FCode = '' then
  begin
    ECode.SetFocus;
    raise Exception.Create('Cdigo Invlido');
  end;

  R := TRegistry.Create;
  try
    if R.OpenKey(Reg, True) then
      R.WriteString('LastCode', FCode);
  finally
    R.Free;
  end;
end;

procedure TForm6.SaveVideoDetails;
var
  R: TRegistry;
  Info: TRegDataInfo;
begin
  if (FCode = '') or (FURL = '') then Exit;

  R := TRegistry.Create;
  try
    if R.OpenKey(Reg+'\'+FCode, True) then
    begin
      R.WriteString('URL', FURL);
      R.WriteString('Name', FName);
      R.WriteInteger('Format', FFormat);
      if R.GetDataInfo('Length', Info) then
        if Info.RegData <> rdBinary then
          R.DeleteValue('Length');
      R.WriteBinaryData('Length', FLength, sizeof(FLength));
    end;
  finally
    R.Free;
  end;
end;

procedure TForm6.VideoWork(ASender: TObject; AWorkMode: TWorkMode; AWorkCount: Int64);
var
  Http: TIdHTTP;
  ContentLength: Int64;
begin
  Http := TIdHTTP(ASender);

  if FLength = 0 then
  begin
    ContentLength := Http.Response.ContentLength;
    FLength := ContentLength;
    SaveVideoDetails;
  end else
    ContentLength := FLength;

  ProgressBar1.Max := ContentLength div 1024;   //KB
  ProgressBar1.Position := (FChunkStart + AWorkCount) div 1024;  // KB

  lProgress.Caption := IntToStr(ProgressBar1.Position div 1024)+'/'+IntToStr(ContentLength div (1024*1024))+' MB';

  Application.ProcessMessages;
end;

end.
